/*
 * Copyright (c) 2012-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#ifndef AGG_TREE_H_
#define AGG_TREE_H_

#include "fabric.h"

#define TREE_PREFIX "tree "
#define TREE_HASH_PREFIX "Tree hash "
#define NODE_PREFIX "node "
#define SUB_NODE_PREFIX "subNode "
#define COMPUTE_PORT_PREFIX "computePort "

typedef std::multimap< uint64_t, class TreeNode * > MapPortKeyToTreeNodes;
//typedef std::map< class AggNode *, uint8_t > MapAggNodeToHostsNumber;
typedef std::vector< sharp_tree_conn > SharpTreeConnVec;
typedef std::vector < SetTreeNodePtr > TreeNodesVec;
typedef std::pair< SetTreeNodePtr::iterator, bool > SetTreeNodePtrInsertRes;

extern const std::size_t g_tree_dump_message_size;

struct JobSubTreeInfo {

    ListTreeNodePtr         m_tree_nodes;
    SharpTreeConnVec        m_tree_connections;
    sharp_tree              m_tree_info;
    AggNodeFabricInfo      *m_tree_root;
    string                  m_log_message;

    void Clear() {
        m_tree_nodes.clear();
        m_tree_connections.clear();
        memset(&m_tree_info, 0, sizeof(m_tree_info));
        m_tree_root = NULL;
        m_log_message.clear();
        m_log_message.reserve(g_tree_dump_message_size);
    }

    void UpdateLogMessage(string &message, int log_level);
    void SetExclusiveLock(bool exclusive_lock);
    void SetDataPathVersion(uint16_t data_path_version_bit_mask);
    void SetExtendedDataTypes(bool extended_data_types_supported);

    int SetPkey(uint16_t pkey) const;
    void GetSubTreeAggNodes(SetAggNodeFabricPtr &agg_nodes) const;
    uint16_t GetMutualAnsNumber(const SetAggNodeFabricPtr &prev_agg_nodes);

    void SetSrcDstPr(ibv_sa_path_rec &path_rec,
                     const PortInfo &src,
                     const PortInfo &dst);

    void SetConnectionPathRec(sharp_tree_conn &connection,
                              const PortInfo &an,
                              const PortInfo &sd,
                              uint16_t pkey,
                              SharpMtu mtu);

    int UpdateTreeConnections(uint32_t connections_number,
                              uint8_t child_index_per_port,
                              uint16_t pkey);

    void ModifyANsAvailableTreeIDs(bool is_available);
    void ModifyANsAvailableTreeIDsLowPriority(bool is_available);

};


struct JobSubTreeScore {

public:

    uint8_t             m_max_radix;
    uint8_t             m_min_radix;
    uint8_t             m_height;
    uint16_t            m_num_ans;
    uint16_t            m_max_load;   //max load on AN
    uint32_t            m_total_load; //load sum of all ANs
    uint32_t            m_path_total_sat_load; //SAT load sum of all AggPaths
    uint16_t            m_max_sat_load;   //max SAT load on AN
    uint32_t            m_total_sat_load; //SAT load sum of all ANs
    uint16_t            m_max_tree_id_sat_load; // max sat load on tree ID
    uint16_t            m_root_sat_load;   // SAT load of the root (not just of the tree, but absolute root)
    uint16_t            m_mutual_ans; //number of mutual ans in multi tree job
    uint16_t            m_data_path_version;

    sharp_quota         m_qota;
    sharp_quota         m_abs_qota; //use requested absolute values

    sharp_percent_t     m_quota_percent;
    sharp_percent_t     m_max_quota_percent;
    sharp_percent_t     m_qps_percent;
    uint8_t             m_job_priority;

    uint8_t             m_child_index_per_port;
    bool                m_multicast_enabled;
    bool                m_allocate_sat;
    bool                m_check_sat;
    bool                m_check_rmc;
    //uint8_t             m_qotaPercentSum;
    //uint8_t             m_qotaPercentMin;
    bool                m_reproducibility_enabled;
    bool                m_exclusive_lock_available;
    uint16_t            m_agg_rate;
    bool                m_fp19_supported;
    bool                m_bfloat19_supported;
    bool                m_rmc_supported;
    bool                m_extended_data_types_supported;
    SharpMtu            m_mtu;
    sharp_resource_priority m_priority;
    uint64_t            m_priority_epoch;

    JobSubTreeScore():m_max_radix(0), m_min_radix(0xFF), m_height(0xFF),
        m_num_ans((uint16_t)-1),
        m_max_load((uint16_t)-1), m_total_load((uint32_t)-1), m_path_total_sat_load(UINT32_MAX),
        m_max_sat_load((uint16_t)-1), m_total_sat_load((uint32_t)-1),
        m_max_tree_id_sat_load((uint16_t)-1),
        m_root_sat_load((uint16_t)-1),
        m_mutual_ans((uint16_t)-1), m_data_path_version(0xFFFF),
        m_quota_percent(0), m_max_quota_percent(0), m_qps_percent(0),
        m_job_priority(0),
        m_child_index_per_port(1),
        m_multicast_enabled(false), m_allocate_sat(false), m_check_sat(false),
        m_check_rmc(false),
        m_reproducibility_enabled(false),
        m_exclusive_lock_available(false),
        m_agg_rate(0xFFFF),
        m_fp19_supported(false),
        m_bfloat19_supported(false),
        m_rmc_supported(false),
        m_extended_data_types_supported(false),
        m_mtu(SHARP_MTU_UNKNOWN),
        m_priority(SHARP_RESOURCE_PRIORITY_NORMAL),
        m_priority_epoch(0)
        //m_qotaPercentSum(0), m_qotaPercentMin(0)
    {
        memset(&m_qota, 0, sizeof(m_qota));
        memset(&m_abs_qota, 0, sizeof(m_abs_qota));
    }

    template <class quota_t> void UpdateQuotaVal(quota_t &quota_val,
                                                 quota_t abs_val,
                                                 quota_t available_quota,
                                                 quota_t total_quota,
                                                 sharp_percent_t default_percent,
                                                 sharp_percent_t max_percent);

    void Reset(const JobResource &job_resource, sharp_resource_priority priority);

    string ToString() const {
        std::stringstream stream;
        stream << "max_radix:" << (int)m_max_radix
            << " min_radix:" << (int)m_min_radix
            << " height:" << (int)m_height
            << " num_ans:" << m_num_ans
            << " max_load:" << m_max_load
            << " total_load:" << m_total_load
            << " max_sat_load:" << m_max_sat_load
            << " total_sat_load:" << m_total_sat_load
            << " max_tree_id_sat_load:" << m_max_sat_load
            << " root_sat_load:" << m_root_sat_load
            << " path_total_sat_load:" << m_path_total_sat_load
            << " mutual_ans:" << m_mutual_ans
            << " data_path_version:" << m_data_path_version
            << " child_index_per_port:" << (uint16_t)m_child_index_per_port
            << " multicast_enabled:" << (m_multicast_enabled ? "true" : "false")
            << " allocate_sat:" << (m_allocate_sat ? "true" : "false")
            << " quota:{q:" << m_qota.max_qps
            << ", ud:" << m_qota.user_data_per_ost
            << ", b:" << m_qota.max_buffers
            << ", o:" << m_qota.max_osts
            << ", g:" << m_qota.max_groups
            << ", rep:" << m_reproducibility_enabled
            << ", exclusive_lock:" << m_exclusive_lock_available
            << ", agg_rate:" << m_agg_rate
            << ", priority:" << m_priority
            << ", priority_epoch" << m_priority_epoch
            << " }";

        return stream.str();
    }

    bool IsWorseThan(const JobSubTreeScore& job_sub_tree_score,
                     JobSubTreeInfo& job_sub_tree_info) const;

    int Update(uint8_t radix,
               uint8_t ports,
               TreeNode &tree_node,
               sharp_trees_t sat_tree_id,
               bool exclusive_lock_available,
               uint16_t agg_rate,
               sharp_resource_priority priority,
               uint32_t agg_path_job_load,
               JobSubTreeInfo &sub_tree_info);

private:

    int CalculateQuota(uint8_t radix,
                       uint8_t ports,
                       uint8_t max_allowed_radix,
                       uint32_t available_buffers,
                       bool enable_reproducibility,
                       const AggNodeInfo &agg_node_info);

    int SetOstsByBuffers(uint8_t radix,
                         uint8_t max_allowed_radix,
                         unsigned int user_data_length,
                         bool enable_reproducibility,
                         const BufferCalculationParameters &buff_calc_param,
                         uint32_t available_buffers);

};

extern const std::size_t g_max_log_message_size;

class AggTree {

    sharp_trees_t           m_tree_id_;
    uint8_t                 m_max_rank_;
    SetTreeNodePtr          m_tree_nodes_;
    bool                    m_is_valid_;
    bool                    m_is_configured_;
    TreeNode               *m_root_tree_node_;
    MapPortKeyToTreeNodes   m_compute_to_tree_nodes_;
    SetPortKey              m_disconnected_compute_ports_;
    SetAnFabricInfoPtr      m_span_leafs; //leaf span by this tree
    AggTree                *m_sat; //SAT tree duplicated from this tree;
    AggTree                *m_llt; //LLT tree duplication source of this tree;
    sharp_resource_priority m_priority_;
    uint64_t                m_tree_hash_;

public:

    AggTree(sharp_trees_t tree_id) :
        m_tree_id_(tree_id),
        m_max_rank_(0), m_is_valid_(true),
        m_is_configured_(false),
        m_root_tree_node_(NULL), m_sat(NULL), m_llt(NULL),
        m_priority_(SHARP_RESOURCE_PRIORITY_NORMAL),
        m_tree_hash_{0} {}

    ~AggTree();

    const SetTreeNodePtr &GetTreeNodes() const { return m_tree_nodes_; }
    sharp_trees_t GetId() const { return m_tree_id_; }
    sharp_resource_priority GetPriority() const { return m_priority_; }
    void SetPriority(sharp_resource_priority priority) { m_priority_ = priority; }

    // Create TreeNode objects for specified AggNodes and set parent-child
    // relation between them.
    //
    // switch_roots enables connecting two sub trees even if p_sub_agg_node is
    // not the root of the second sub tree.
    // When switch_roots is enabled, the second sub tree is rotated to make
    // p_sub_agg_node its root, and then connects it as a child of p_agg_node.
    //
    // This method modifies AggNode objects.
    int CreateTreeConnection(AggNodeFabricInfo *p_agg_node,
                             AggNodeFabricInfo *p_sub_agg_node,
                             bool switch_roots = false);

    //create ree node with no Descendant
    int CreateTreeNode(AggNodeFabricInfo *p_agg_node);

    int ConfigureOnFabric();
    int RemoveConfigurationFabric();

    int SetRoot();
    int SetRoot(AggNodeFabricInfo *p_agg_node);
    int SetTreeNodesRank();

    int DuplicateTree(const AggTree *p_src_tree);
    TreeNode *DuplicateSubTree(TreeNode *p_src_node);
    TreeNode *GetRootTreeNode() { return m_root_tree_node_; };
    void SetRootTreeNode(TreeNode *tree_node) { m_root_tree_node_ = tree_node; }

    bool IsValid()const { return m_is_valid_; }
    bool IsConfigured()const { return m_is_configured_; }

    bool IsNodesSupportsSat();

    int UpdateComputeToTreeNode(
        MapPortKeyToAggNodeFabrics &compute_por_to_agg_nodes,
        bool skip_if_exist);

    int GetJobSubTreeInfo(const SetPortDataConstPtr& compute_ports,
                          const JobResource &job_resource,
                          JobSubTreeInfo &job_sub_tree_info,
                          const JobSubTreeScore &best_result,
                          JobSubTreeScore &sub_tree_result);

    int PrepareSubTreeForJob(JobSubTreeInfo &job_sub_tree_info);

    int UpdateTreeTree();
    int UpdateDfpTree(const SetAnFabricInfoPtr &span_leafs_diff);
    int BuildTree(AggNodeFabricInfo *p_tree_root,
                  const SetAnFabricInfoPtr &descendant_leafs,
                  const AggNodeFabricInfo *parent_agg_node);

    void DumpTree(std::string &dump_message) const;
    void DumpState(FILE *f) const;

    bool IsComputePortMappedtToTreeNode(uint64_t port_key) const
    { return m_compute_to_tree_nodes_.end() != m_compute_to_tree_nodes_.find(port_key); }

    void ClearComputeToTreeNode(port_key_t compute_port_key) {
        m_compute_to_tree_nodes_.erase(compute_port_key);
        m_disconnected_compute_ports_.erase(compute_port_key);
    }

    AggTree *GetSat() { return m_sat; }
    void SetSat(AggTree *p_sat) { m_sat = p_sat; }

    AggTree *GetLlt() { return m_llt; }
    void SetLlt(AggTree *p_llt) { m_llt = p_llt; }
    bool IsSat()const { return m_llt != NULL; }

    int SelectMostFreeAggPaths();

    inline uint64_t GetHash() const { return m_tree_hash_; };
    inline void SetHash(const uint64_t new_tree_hash) { m_tree_hash_ = new_tree_hash; };
    void CalculateHash();


private:

    int UpdateTreeNodesVec(TreeNodesVec &tree_nodes_vec,
                           uint8_t &min_tree_rank,
                           uint8_t &max_tree_rank,
                           TreeNode *p_tree_node,
                           const PortData *p_compute_port,
                           uint8_t child_index,
                           JobSubTreeScore &sub_tree_result,
                           JobSubTreeInfo &sub_tree_info);

    // return true if lock available, false otherwise.
    bool CheckExclusiveLockAvailability(TreeNode *p_tree_node, bool is_sat_job, bool is_root=false);
    bool IsLockAvailable(TreeNode *p_tree_node, bool is_sat_job, bool is_root=false);

    static void GetTreeNodesList(TreeNodesVec &tree_nodes_vec,
                                 ListTreeNodePtr &tree_nodes);

    // Rotate subtree untill specified TreeNode object becomes root
    int RotateSubTree(TreeNode *p_tree_node);

    // During BuildTree span all non leaf ANs if required
    int SpanUnconnectedDescendants(AggNodeFabricInfo *p_sub_root_fabric_info);

};

#endif //AGG_TREE_H_
